﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CatanGame.Model
{
    public class CatanBoard
    {
        #region Private fields
        private readonly static Dictionary<ResourceType, int> resourcePossibilities = new Dictionary<ResourceType, int> {
            { ResourceType.None, 1 },
            { ResourceType.Grain, 4 },
            { ResourceType.Lumber, 4 },
            { ResourceType.Wool, 4 },
            { ResourceType.Brick, 3 },
            { ResourceType.Ore, 3 },
        };

        private readonly static Dictionary<int, int> numPossibilities = new Dictionary<int, int>
        {
            { 2, 1 },
            { 3, 2 },
            { 4, 2 },
            { 5, 2 },
            { 6, 2 },
            { 8, 2 },
            { 9, 2 },
            { 10, 2 },
            { 11, 2 },
            { 12, 1 },
        };

        private Player[] _players;
        private NodeMatrix _nodes;
        private RoadMatrix _roads;
        private Field[,] _fields;
        private List<Field>[] _fieldNumbers;
        private int _currentPlayer;
        private static Random rand = new Random();
        private bool _initialTurn;
        private int _turnIncrement;
        private bool _rolled;
        #endregion

        #region Public properties
        public Player[] Players { get => _players; }
        public NodeMatrix Nodes { get => _nodes; }
        public RoadMatrix Roads { get => _roads; }
        public Field[,] Fields { get => _fields; }
        public (int, int) RobberPosition { get; set; }
        public Player LongestRoadOwner { get; set; }
        public Player LargestArmyOwner { get; set; }
        public bool IsGameOver { get; set; }
        public int Size { get; set; }

        public bool InitialTurn { get => _initialTurn; }

        public int CurrentPlayerId { get => _currentPlayer; }

        public int HexagonCount { get => Size * Size - (Size / 2 + 1) * Size / 2; }

        public int VertexCount { get => HexagonCount * 2 + Size * 3 + 1; }
        #endregion

        #region Public constructors
        public CatanBoard(int size = 7)
        {
            if (size % 2 == 0)
                throw new InvalidOperationException("size can't be an even number!");
            Size = size;

            InitializeNodes();
            InitializeFields();
            InitializeRoads();
            InitializePlayers();

        }
        #endregion

        #region Public methods

        public void BuyCard(Player p)
        {
            if (p.CanBuyDevelopmentCard)
            {
                int cardType = rand.Next(0, 2);
                p.DevelopmentCards[(CardType)cardType] += 1;
                if ((CardType)cardType == CardType.VictoryPoint)
                    p.Score++;
                p.DevelopmentCardBought();
            }
        }

        public bool PlaceRoad(RoadCoordinates coords, Player p)
        {
            if (!CanPlaceRoad(coords, p))
                return false;

            _roads.Add(coords, p.Id);
            p.RoadBuilt(_initialTurn);

            if (_initialTurn)
                StepGame();

            return true;
        }

        public List<RoadCoordinates> EnabledRoads(Player p)
        {
            List<RoadCoordinates> answer = new();
            if (_rolled)
            for (int r = 0; r < Size; r++)
                for (int q = 0; q < Size; q++)
                    for (int d = 0; d < 2; d++)
                        foreach (RoadCoordinates road in EnabledRoadsFromNode(p, new NodeCoordinates(r, q, (Direction)d)))
                            answer.Add(road);
                    
            return answer;
        }

        public List<NodeCoordinates> EnableNodes(Player p)
        {
            List<NodeCoordinates> answer = new();
            if (_rolled)
            for (int r = 0; r < Size; r++)
                for (int q = 0; q < Size; q++)
                    for (int d = 0; d < 2; d++)
                        if (CanPlaceSettlement(new NodeCoordinates(r, q, (Direction)d), p))
                            answer.Add(new NodeCoordinates(r, q, (Direction)d));
            return answer;
        }

        public List<RoadCoordinates> EnabledRoadsFromNode(Player p, NodeCoordinates current)
        {
            List<RoadCoordinates> answer = new List<RoadCoordinates>();
            List<NodeCoordinates> neighbors = GetNeighborCoords(current);
            foreach (NodeCoordinates neighbor in neighbors)
            {
                    // both nodes must exist in order to place a road on them
                    if (Nodes.NodeExists(current) && Nodes.NodeExists(neighbor))
                    {
                        RoadCoordinates roadToPlace = new RoadCoordinates(current, neighbor);
                        if (!_roads.IsPlacedRoad(roadToPlace) && CanPlaceRoad(roadToPlace, p))
                            answer.Add(roadToPlace);
                    }
            }
            return answer;
        }

        public bool PlaceSettlement(NodeCoordinates node, Player p)
        {
            if (CanPlaceSettlement(node, p))
            {
                House building = Nodes[node].Building;
                building.Type = HouseType.Settlement;
                building.Owner = p;
                p.AddedSettlement(_initialTurn);
                if (!_initialTurn)
                    Players[CurrentPlayerId].Score++;
                return true;
            }
            return false;
        }

        public bool UpgradeToCity(NodeCoordinates node, Player p)
        {
            if (CanUpgradeToCity(node, p))
            {
                Nodes[node].Building.Type = HouseType.City;
                p.UpgradedToCity();
                Players[CurrentPlayerId].Score++;
                return true;
            }
            return false;
        }

        public void StepGame()
        {
            if (_rolled || _initialTurn)
                _currentPlayer += _turnIncrement;
            if (!_initialTurn)
                _rolled = false;
            if (_currentPlayer == _players.Length)
            {
                if (_initialTurn)
                {
                    _turnIncrement *= -1;
                    _currentPlayer += _turnIncrement;
                }
                else
                    _currentPlayer = 0;
            } 
            if (_currentPlayer == -1)
            {
                _initialTurn = false;
                _rolled = false;
                _turnIncrement *= -1;
                _currentPlayer += _turnIncrement;
            }
        }

        public bool Trade(ResourceType type1, ResourceType type2, int cnt)
        {
            if (_players[_currentPlayer].Resources[type1] >= 3 * cnt)
            {
                _players[_currentPlayer].Resources[type1] -= 3 * cnt;
                _players[_currentPlayer].Resources[type2] += cnt;
                return true;
            }
            else
                return false;
        }

        public void MoveRobber((int r, int q) coords)
        {
            RobberPosition = (coords.r, coords.q);
        }

        public (int, int) Roll()
        {
            _rolled = true;
            return (rand.Next(1, 7), rand.Next(1, 7));
        }

        public void SendResources(int num)
        {
            for (int r = 0; r < Size; r++)
                for (int q = 0; q < Size; q++)
                    if (_fields[r, q] != null &&_fields[r,q].FieldNum == num && (r, q) != RobberPosition)
                        _fields[r, q].SendResources();
        }

        public bool CanPlaceRoad(RoadCoordinates coords, Player p)
        {
            bool answer = false;

            if (_initialTurn)
            {
                if (Nodes.NodeExists(coords.From) && Nodes[coords.From].Building.HasOwner)
                    answer |= Nodes[coords.From].Building.Owner.Id == p.Id;
                if (Nodes.NodeExists(coords.To) && Nodes[coords.To].Building.HasOwner)
                    answer |= Nodes[coords.To].Building.Owner.Id == p.Id;

                if (answer)
                    return true;
            }

            Predicate<Road> isConnected = (neighboringRoad) =>
                    neighboringRoad.PlayerId == p.Id
                    && (Nodes[neighboringRoad.Coordinates.From].Building.Type == HouseType.NoBuilding
                        || Nodes[neighboringRoad.Coordinates.From].Building.Owner.Id == p.Id);
            
            answer |= Roads.RoadsFromNode(coords.From).Exists(isConnected)
                   || Roads.RoadsFromNode(coords.To).Exists(isConnected);

            return answer;
        }

        public bool CanPlaceSettlement(NodeCoordinates node, Player p)
        {
            if (_currentPlayer != p.Id)
                return false;

            List<NodeCoordinates> neighbors = GetNeighborCoords(node);
            bool answer = Nodes.NodeExists(node) 
                && Nodes[node].Building.Type == HouseType.NoBuilding 
                && p.CanBuildSettlement;

            answer &= neighbors.All((neighbor) =>
                !Nodes.NodeExists(neighbor) 
                || Nodes[neighbor].Building.Type == HouseType.NoBuilding);

            bool hasNeighboringRoad = Roads.RoadsFromNode(node).Any((road) => road.PlayerId == p.Id);

            answer &= hasNeighboringRoad || _initialTurn;

            return answer;
        }

        public bool CanUpgradeToCity(NodeCoordinates node, Player p)
        {
            return Nodes[node].Building.Owner == p && p.CanBuildCity;
        }

        #endregion

        #region Private methods
        private void InitializeFields()
        {
            _fields = new Field[Size, Size];
            Dictionary<ResourceType, int> possibleResources = new(resourcePossibilities);
            Dictionary<int, int> possibleNums = new(numPossibilities);

            int innerBoardSize = Size - 2;
            for(int q = 0; q < Size; ++q)
            {
                for (int r = 0; r < Size; ++r)
                {
                    if (Math.Abs(q + r - Size + 1) < 1 + Size / 2) 
                    {
                        if (IsInnerBoard(r, q))
                        {
                            _fields[r, q] = new Field(new List<Node> {
                                Nodes[r,     q,      Direction.Bottom],
                                Nodes[r,     q,      Direction.Top],
                                Nodes[r - 1, q + 1,  Direction.Bottom],
                                Nodes[r - 1, q,      Direction.Bottom],
                                Nodes[r + 1, q - 1,  Direction.Top],
                                Nodes[r + 1, q,      Direction.Top],
                            });
                            _fields[r, q].ResourceType = GetNextResourceType(possibleResources);
                            if (_fields[r, q].ResourceType != ResourceType.None)
                                _fields[r, q].FieldNum = GetNextFieldNum(possibleNums);
                        }
                        else
                        {
                            _fields[r, q] = new Field();
                            _fields[r, q].ResourceType = ResourceType.Sea;
                        }

                        if (_fields[r, q].ResourceType == ResourceType.None)
                            RobberPosition = (r, q);
                    }
                }
            }
        }

        private bool IsInnerBoard(int r, int q)
        {
           return Math.Abs(q - Size / 2 + r - Size / 2) < 1 + (Size - 2) / 2 && q != 0 && q != Size - 1 && r != 0 && r != Size - 1;
        }

        private bool AreValidCoordinates(NodeCoordinates coords)
        {
            return coords.R >= 0 && coords.Q >= 0 && coords.R < Size && coords.Q < Size;
        }

        private void InitializeNodes()
        {
            _nodes = new NodeMatrix(Size);
        }



        private void InitializeRoads()
        {
            _roads = new RoadMatrix(Size);
        }

        private void InitializePlayers()
        {
            _currentPlayer = 0;
            _turnIncrement = 1;
            _initialTurn = true;
            _rolled = true;
            _players = new Player[3];
            for (int i = 0; i < _players.Length; ++i)
            {
                _players[i] = new Player();
                _players[i].Id = (byte) i;
            }
        }

        private ResourceType GetNextResourceType(Dictionary<ResourceType, int> possibleResources)
        {
            List<ResourceType> possibilitiesList = possibleResources.Keys.ToList();
            ResourceType randResource = possibilitiesList[rand.Next(possibilitiesList.Count)];
            --possibleResources[randResource];

            if (possibleResources[randResource] == 0)
                possibleResources.Remove(randResource);

            return randResource;
        }
        private int GetNextFieldNum(Dictionary<int, int> possibleResources)
        {
            List<int> possibilitiesList = possibleResources.Keys.ToList();
            int randNum = possibilitiesList[rand.Next(possibilitiesList.Count)];
            --possibleResources[randNum];

            if (possibleResources[randNum] == 0)
                possibleResources.Remove(randNum);

            return randNum;
        }



        private List<NodeCoordinates> GetNeighborCoords(NodeCoordinates node)
        {
            List<NodeCoordinates> answer = new();
            switch (node.Dir)
            {
                case Direction.Top:
                    {
                        if (node.R > 0)
                            answer.Add(new NodeCoordinates(node.R - 1, node.Q, Direction.Bottom));
                        if (node.R > 0 && node.Q < 6)
                            answer.Add(new NodeCoordinates(node.R - 1, node.Q + 1, Direction.Bottom));
                        if (node.R > 1 && node.Q < 6)
                            answer.Add(new NodeCoordinates(node.R - 2, node.Q + 1, Direction.Bottom));
                            //return new List<NodeCoordinates> { new NodeCoordinates(node.R - 1, node.Q    , Direction.Bottom),
                            //                                   new NodeCoordinates(node.R - 1, node.Q + 1, Direction.Bottom),
                            //                                   new NodeCoordinates(node.R - 2, node.Q + 1, Direction.Bottom)};
                        return answer;
                    }
                case Direction.Bottom:
                    {
                        if (node.R < 6)
                            answer.Add(new NodeCoordinates(node.R + 1, node.Q    , Direction.Top));
                        if (node.Q > 0 && node.R < 6)
                            answer.Add(new NodeCoordinates(node.R + 1, node.Q - 1, Direction.Top));
                        if (node.R < 5 && node.Q > 0)
                            answer.Add(new NodeCoordinates(node.R + 2, node.Q - 1, Direction.Top));
                        //return new List<NodeCoordinates> { new NodeCoordinates(node.R + 1, node.Q - 1, Direction.Top),
                        //                                   new NodeCoordinates(node.R + 1, node.Q    , Direction.Top),
                        //                                   new NodeCoordinates(node.R + 2, node.Q - 1, Direction.Top)};
                        return answer;
                    }
                default:
                    throw new InvalidOperationException("Invalid Node direction!");
            }
        }
        #endregion

    }
}
