﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using CatanGame.Model;

namespace CatanGame.ViewModel
{
    public class CatanViewModel : ViewModelBase
    {
        #region Private fields
        private readonly GameModel _model;

        private int _decies1 ;
        private int _decies2;
        private PlayerViewModel _currentPlayer;
        private PlayerViewModel _redPlayer;
        private PlayerViewModel _greenPlayer;
        private PlayerViewModel _bluePlayer;
        #endregion

        #region Public properties
        public ObservableCollection<FieldViewModel> Fields { get; set; }
        public ObservableCollection<ObservableCollection<RoadViewModel>> Roads { get; set; }
        public ObservableCollection<NodeViewModel> Nodes { get; set; }
        public ObservableCollection<PlayerViewModel> Players { get; set; }
        public ResourceType[] Resources { get; set; }
        public PlayerViewModel CurrentPlayer { get => _currentPlayer; set => ChangeProperty(ref _currentPlayer, value); }
        public int BoardSize { get => _model.Board.Size; }
        public int NodeCount { get => _model.Board.VertexCount; }
        public int CurrentPlayerId { get => _model.CurrentPlayerId; }
        #endregion

        #region Events
        public event EventHandler<RollEventArgs> GameRolled;
        public event EventHandler NewGameView;
        #endregion

        #region Public constructors

        public CatanViewModel(GameModel model)
        {
            //model csatlakozatatasa
            _model = model;
            _model.NodeUpgraded += Model_NodeChanged;
            _model.RoadPlaced += Model_RoadPlaced;
            _model.RoadsEnabled += Model_RoadsEnabled;
            _model.NodesEnabled += Model_NodesEnabled;
            _model.GameStepped += new EventHandler(Model_GameStepped);
            _model.GameRolled += Model_Rolled;

            NewGame();
        }



        #endregion

        #region Private methods
        private void NewGame()
        {

            _model.NewGame();

            Init();
        }
        private void Init()
        {
            // initialize commands
            ShowEnabledRoadsCommand = new DelegateCommand(ShowEnabledRoads);
            ShowEnabledNodesCommand = new DelegateCommand(ShowEnabledNodes);
            StepGameCommand = new DelegateCommand(GameStep);
            RollCommand = new DelegateCommand(GameRoll);
            NewGameViewCommand = new DelegateCommand(param => { OnNewGameView(); });

            // initialize collections
            InitPlayers();
            InitFields();
            InitNodes();
            InitRoads();
            ShowEnabledNodes(null);
        }

        private void InitPlayers()
        {
            Players = new();
            PlayerViewModel player = null;
            for (int i = 0; i < _model.Board.Players.Length; i++)
            {
                player = new PlayerViewModel()
                {
                    Id = _model.Board.Players[i].Id,
                    Score = _model.Board.Players[i].Score,
                    LumberCnt = _model.Board.Players[i].Resources[ResourceType.Lumber],
                    BrickCnt = _model.Board.Players[i].Resources[ResourceType.Brick],
                    WoolCnt = _model.Board.Players[i].Resources[ResourceType.Wool],
                    GrainCnt = _model.Board.Players[i].Resources[ResourceType.Grain],
                    OreCnt = _model.Board.Players[i].Resources[ResourceType.Ore],
                };
                Players.Add(player);
            }
            CurrentPlayer = Players[CurrentPlayerId];
        }

        private void InitFields()
        {
            Fields = new();
            FieldViewModel field = null;
            for (int r = 0; r < _model.Board.Size; ++r)
            {
                for (int q = 0; q < _model.Board.Size; ++q)
                {
                    field = null;
                    if (_model.Board.Fields[r, q] is not null)
                    {
                        field = new FieldViewModel()
                        {
                            R = r,
                            Q = q,
                            FieldNumber = _model.Board.Fields[r, q].FieldNum,
                            FieldType = _model.Board.Fields[r, q].ResourceType.ToString(),
                            HasRobber = false,
                        };

                    }
                    Fields.Add(field);
                }
            }
        }

        private void InitNodes()
        {
            Nodes = new();
            for (int r = 0; r < _model.Board.Size; ++r)
            {
                for (int q = 0; q < _model.Board.Size; ++q)
                {
                    for (int dir = 0; dir < 2; ++dir)
                    {
                        NodeViewModel node = null;
                        NodeCoordinates coords = new NodeCoordinates(r, q, (Direction)dir);
                        Node boardNode = _model.Board.Nodes[coords];
                        if (boardNode is not null)
                        {
                            node = new NodeViewModel()
                            {
                                //HouseOwner = boardNode.Building.Owner.Name,
                                HouseType = boardNode.Building.Type.ToString(),
                                Position = (r, q, dir),
                                UpgradeNodeCommand = new DelegateCommand((param) => UpgradeNode?.Invoke(this, coords)),
                                HouseOwner = "",
                                HouseOwnerId = -1
                            };

                        }
                        Nodes.Add(node);
                    }
                }
            }
        }
  
        private void InitRoads()
        {
            int totalNodeCount = _model.Board.Size * _model.Board.Size * 2;
            Roads = new();
            // Note: iterating vertically to make symmetric matrix share references of Roads
            for (int i = 0; i < totalNodeCount; ++i)
                Roads.Add(new());

            for (int j = 0; j < totalNodeCount; ++j)
            {
                for (int i = 0; i < totalNodeCount; ++i)
                {
                    if (j > i)
                        Roads[i].Add(Roads[j][i]);
                    else
                    {
                        int iCop = i;
                        int jCop = j;
                        Roads[i].Add(new RoadViewModel()
                        {
                            IsBuildable = false,
                            Position = (iCop, jCop),
                            BuildRoadCommand = new DelegateCommand((param) => PlaceRoad?.Invoke(this, (iCop, jCop))),
                            OwnerId = -1,
                        });
                    }
                }
            }
        }

        private void RefreshDatas()
        {
            CurrentPlayer.Score = _model.Board.Players[CurrentPlayerId].Score;
            CurrentPlayer.LumberCnt = _model.Board.Players[CurrentPlayerId].Resources[ResourceType.Lumber];
            CurrentPlayer.BrickCnt = _model.Board.Players[CurrentPlayerId].Resources[ResourceType.Brick];
            CurrentPlayer.WoolCnt = _model.Board.Players[CurrentPlayerId].Resources[ResourceType.Wool];
            CurrentPlayer.GrainCnt = _model.Board.Players[CurrentPlayerId].Resources[ResourceType.Grain];
            CurrentPlayer.OreCnt = _model.Board.Players[CurrentPlayerId].Resources[ResourceType.Ore];
        }

        private void ShowEnabledRoads(object param)
        {
            _model.SendEnabledRoads(CurrentPlayerId);
        }
        private void ShowEnabledNodes(object param)
        {
            _model.SendEnabledNodes(CurrentPlayerId);
        }

        private void GameStep(object param)
        {
            _model.StepGame();
        }
        private void GameRoll(object param)
        {
            
            (_decies1,_decies2)=_model.Roll();
        }

        private int GetNodeId(int r, int q, int dir)
        {
            return r * BoardSize * 2 + q * 2 + dir;
        }

        private int GetNodeId((int R, int Q, int Dir) coords)
        {
            return GetNodeId(coords.R, coords.Q, coords.Dir);
        }

        private int GetNodeId(NodeCoordinates coords)
        {
            return GetNodeId(coords.R, coords.Q, (int) coords.Dir);
        }

        #endregion

        #region Public commands
        public DelegateCommand NewGameCommand { get; private set; }
        public DelegateCommand NewGameViewCommand { get; private set; }
        public DelegateCommand ExitCommand { get; private set; }
        public DelegateCommand BuyCardCommand { get; private set; }
        public DelegateCommand PlaceRoadCommand { get; private set; }
        public DelegateCommand PlaceSettlementCommand { get; private set; }
        public DelegateCommand UpgradeToCityCommand { get; private set; }
        public DelegateCommand StepGameCommand { get; private set; }
        public DelegateCommand RollCommand { get; private set; }
        public DelegateCommand ShowEnabledRoadsCommand { get; private set; }
        public DelegateCommand ShowEnabledNodesCommand { get; private set; }
        #endregion

        #region Public methods
        #endregion


        #region Private methods

        private void ClearOptions()
        {
            ClearBuildableRoads();
            ClearBuildableNodes();
        }

        private void ClearBuildableRoads()
        {
            foreach (var seconds in Roads)
                foreach (var road in seconds)
                    if (road is not null)
                        road.IsBuildable = false;
        }

        private void ClearBuildableNodes()
        {
            foreach (NodeViewModel node in Nodes)
                if (node is not null)
                    node.IsBuildable = false;
        }

        #endregion

        #region Model event handlers
        private void Model_CardBought(object obj, EventArgs e)
        {

        }


        private void Model_NodeChanged(object obj, NodeCoordinates coords)
        {
            int nodeId = coords.R * BoardSize * 2 + coords.Q * 2 + (int)coords.Dir;
            Node node = _model.Board.Nodes[coords];

            Nodes[nodeId].HouseType = node.Building.Type.ToString();
            Nodes[nodeId].HouseOwnerId = node.Building.Owner.Id;

            RefreshDatas();

            ClearOptions();
        }

        private void Model_RoadPlaced(object obj, RoadPlacedEventArgs eventArgs)
        {
            Roads[GetNodeId(eventArgs.FirstNode)][GetNodeId(eventArgs.SecondNode)].OwnerId = eventArgs.OwnerId;

            RefreshDatas();

            ClearOptions();
        }
        private void Model_GameStepped(object obj, EventArgs e)
        {
            CurrentPlayer = Players[_model.CurrentPlayerId];
            ClearOptions();
        }
        private void Model_Rolled(object obj, RollEventArgs e)
        {
            GameRolled?.Invoke(this, new RollEventArgs(e.Decies1,e.Decies2));
            RefreshDatas();
        }
        private void OnNewGameView()
        {
            NewGameView?.Invoke(this, EventArgs.Empty);
        }

        private void Model_RoadsEnabled(object obj, RoadsEnabledEventArgs eventArgs)
        {
            ClearOptions();

            // setting IsBuildable to true at the reverse coordinates is not needed, because 
            // the normal coordinates and the reverse coordinates share the same reference.
            eventArgs.EnabledRoads.ForEach((road) => {
                RoadViewModel roadVM = Roads[GetNodeId(road.From)][GetNodeId(road.To)];
                if (roadVM is not null)
                   roadVM.IsBuildable = true;
            });
        }

        private void Model_NodesEnabled(object obj, List<NodeCoordinates> enabledNodes)
        {
            ClearOptions();

            enabledNodes.ForEach((node) => Nodes[GetNodeId(node)].IsBuildable = true);
        }


        #endregion

        #region Public events

        //public event EventHandler<EventArgs> NewGame;
        public event EventHandler<EventArgs> Exit;
        public event EventHandler<EventArgs> BuyCard;
        public event EventHandler<(int FirstNodeId, int SecondNodeId)> PlaceRoad;
        public event EventHandler<NodeCoordinates> UpgradeNode;
        public event EventHandler<EventArgs> UpgradeToCity;
        public event EventHandler<EventArgs> StepGame;
        public event EventHandler<EventArgs> Roll;

        #endregion
    }
}
